// API - HPDI32 - Any OS
// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/HPDI32/HPDI32_Linux_sdk_6.x.x_GSC_DN/api/dev_access.c $
// $Rev: 23674 $
// $Date: 2013-10-31 16:25:34 -0500 (Thu, 31 Oct 2013) $

#include <stdio.h>
#include <string.h>

#include "dev_access.h"

#include "gsc_version.h"
#include "gsc_drv.h"
#include "gsc_reg.h"
#include "gsc_sleep.h"

#include "gsc_os_init.h"

#include "dev_config.h"
#include "dev_irq.h"
#include "dev_reg.h"
#include "dev_version.h"



/* variables	*************************************************************/

static	dev_data_t	_dev_list[GSC_PROCESS_OPEN_MAX];
static	gsc_sem_t	_dev_sem;



/*****************************************************************************
*
*	Function:	_app_handle_validate
*
*	Purpose:
*
*		Validate the handle given by the application and provide the device
*		data structure pointer, if applicable.
*
*	Arguments:
*
*		handle		The handle to validate.
*
*		dev			The structure pointer is recorded here.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
*****************************************************************************/

static U32 _app_handle_validate(void* handle, dev_data_t** dev)
{
	U8	lst_index;
	U32	status;

	status	= gsc_handle_validate(	handle,
									ARRAY_ELEMENTS(_dev_list),
									sizeof(_dev_list[0]),
									&_dev_list[0].device,
									&lst_index);

	if (status == (U32) GSC_SUCCESS)
		dev[0]	= &_dev_list[lst_index];
	else
		dev[0]	= NULL;

	return(status);
}



/*****************************************************************************
*
*	Function:	_dev_access_lock
*
*	Purpose:
*
*		Get exclusive access to the given device.
*
*	Arguments:
*
*		dev			The device to lock.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
*****************************************************************************/

static U32 _dev_access_lock(dev_data_t* dev)
{
	U32	status;

	status	= gsc_sem_lock(&dev->sem);
	return(status);
}



/*****************************************************************************
*
*	Function:	_dev_data_create
*
*	Purpose:
*
*		Prepare the given structure for initial use with the respective
*		device. This function is primarily for items which do not change
*		while the device is open.
*
*	Arguments:
*
*		dev			The data structure to initialize.
*					We retain the device substructure.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
*****************************************************************************/

static U32 _dev_data_create(dev_data_t* dev)
{
	static const char*	irq_name[]	=
	{
		/* 0	*/	"FVS",
		/* 1	*/	"FVE",
		/* 2	*/	"C1:LV/GPIO0",
		/* 3	*/	"C2:SV/GPIO1",
		/* 4	*/	"C3:RxR/GPIO2",
		/* 5	*/	"C4:TxR/GPIO3",
		/* 6	*/	"C5:TxE/GPIO4",
		/* 7	*/	"C6:RxE/GPIO5",
		/* 8	*/	"Tx FIFO E",
		/* 9	*/	"Tx FIFO AE",
		/* 10	*/	"Tx FIFO AF",
		/* 11	*/	"Tx FIFO F",
		/* 12	*/	"Rx FIFO E",
		/* 13	*/	"Rx FIFO AE",
		/* 14	*/	"Rx FIFO AF",
		/* 15	*/	"Rx FIFO F"
	};

	char			buf[64];
	gsc_device_t	device;
	int				i;
	U32				stat;
	U32				status	= (U32) GSC_SUCCESS;

	device		= dev->device;
	memset(dev, 0, sizeof(dev[0]));
	dev->device	= device;

	stat	= gsc_sem_create(&dev->sem, "DEV");
	status	= (status == (U32) GSC_SUCCESS) ? stat : status;

	stat	= gsc_bar_create(&dev->plx_regs, &dev->device, 0, GSC_PCI_9080_BAR0);
	status	= (status == (U32) GSC_SUCCESS) ? stat : status;
	stat	= gsc_bar_create(&dev->gsc_regs, &dev->device, 2, GSC_PCI_9080_BAR2);
	status	= (status == (U32) GSC_SUCCESS) ? stat : status;

	stat	= gsc_dma_create(&dev->dma, &dev->device);
	status	= (status == (U32) GSC_SUCCESS) ? stat : status;

	stat	= gsc_io_create(&dev->rx.gio,
							"I/O Rx",
							&dev->device,
							dev,
							&dev->rx,
							&dev->gsc_regs,
							&dev->dma,
							0);
	status	= (status == (U32) GSC_SUCCESS) ? stat : status;
	stat	= gsc_io_create(&dev->tx.gio,
							"I/O Tx",
							&dev->device,
							dev,
							&dev->tx,
							&dev->gsc_regs,
							&dev->dma,
							1);
	status	= (status == (U32) GSC_SUCCESS) ? stat : status;

	stat	= gsc_mem_create(&dev->rx.mem, "I/O Rx mem", &dev->device);
	status	= (status == (U32) GSC_SUCCESS) ? stat : status;
	stat	= gsc_mem_create(&dev->tx.mem, "I/O Tx mem", &dev->device);
	status	= (status == (U32) GSC_SUCCESS) ? stat : status;

	/* Interrupt Notification */
	stat	= gsc_irq_create(&dev->irq.girq, &dev->device);
	status	= (status == (U32) GSC_SUCCESS) ? stat : status;

	stat	= gsc_irq_main_create(	&dev->irq.girq,
									"IRQ DMA 0",
									GSC_MAIN_INT_DMA_0,
									&dev->dma.ch0.irq);
	status	= (status == (U32) GSC_SUCCESS) ? stat : status;

	stat	= gsc_irq_main_create(	&dev->irq.girq,
									"IRQ DMA 1",
									GSC_MAIN_INT_DMA_1,
									&dev->dma.ch1.irq);
	status	= (status == (U32) GSC_SUCCESS) ? stat : status;

	stat	= gsc_irq_main_create(	&dev->irq.girq,
									"IRQ Local",
									GSC_MAIN_INT_LOCAL,
									&dev->irq.local);
	status	= (status == (U32) GSC_SUCCESS) ? stat : status;

	for (i = 0; i < (int) ARRAY_ELEMENTS(dev->irq.locals); i++)
	{
		sprintf(buf, "IRQ Local %s", irq_name[i]);
		stat	= gsc_irq_local_create(	&dev->irq.girq,
										buf,
										(U32) i,
										&dev->irq.locals[i]);
		status	= (status == (U32) GSC_SUCCESS) ? stat : status;
	}

	return(status);
}



/*****************************************************************************
*
*	Function:	_dev_data_destroy
*
*	Purpose:
*
*		Close the given data structure as the referenced device in no longer
*		being used.
*
*	Arguments:
*
*		dev			The data structure to initialize.
*					We retain the device substructure.
*
*	Returned:
*
*		None.
*
*****************************************************************************/

static void _dev_data_destroy(dev_data_t* dev)
{
	gsc_device_t	device;

	gsc_io_destroy(&dev->rx.gio);
	gsc_io_destroy(&dev->tx.gio);

	gsc_mem_destroy(&dev->tx.mem);
	gsc_mem_destroy(&dev->rx.mem);

	gsc_dma_destroy(&dev->dma);

	gsc_irq_destroy(&dev->irq.girq);

	gsc_bar_destroy(&dev->plx_regs);
	gsc_bar_destroy(&dev->gsc_regs);

	gsc_sem_destroy(&dev->sem);

	device		= dev->device;
	memset(dev, 0, sizeof(dev[0]));
	dev->device	= device;
}



/*****************************************************************************
*
*	Function:	_dev_reset
*
*	Purpose:
*
*		Reset the HPDI32.
*
*	Arguments:
*
*		dev			The device to reset.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
*****************************************************************************/

static U32 _dev_reset(dev_data_t* dev)
{
	U32	stat;
	U32	status	= (U32) GSC_SUCCESS;

	stat	= io_abort(dev, HPDI32_WHICH_RX, HPDI32_IO_ABORT_YES, NULL);
	status	= (status == (U32) GSC_SUCCESS) ? stat : status;
	stat	= io_abort(dev, HPDI32_WHICH_TX, HPDI32_IO_ABORT_YES, NULL);
	status	= (status == (U32) GSC_SUCCESS) ? stat : status;

	// Put the board in the default state we need it in.
	stat	= GSC_REG_WRITE(dev, HPDI32_BCR, HPDI32_BCR_BOARD_RESET);
	gsc_sleep_ms(1);	// Sleep for 1ms though a reset takes only 10us.
	stat	= GSC_REG_WRITE(dev, HPDI32_BCR, HPDI32_BCR_DEFAULT);
	status	= (status == (U32) GSC_SUCCESS) ? stat : status;

	if (dev->cache.fsr & HPDI32_FSR_IELR_IHLR)
	{
		stat	= GSC_REG_WRITE(dev, HPDI32_IELR, HPDI32_IELR_DEFAULT);
		status	= (status == (U32) GSC_SUCCESS) ? stat : status;
		stat	= GSC_REG_WRITE(dev, HPDI32_IHLR, HPDI32_IHLR_DEFAULT);
		status	= (status == (U32) GSC_SUCCESS) ? stat : status;
	}

	stat	= GSC_REG_WRITE(dev, HPDI32_ICR, HPDI32_ICR_DEFAULT);
	status	= (status == (U32) GSC_SUCCESS) ? stat : status;

	return(status);
}



/*****************************************************************************
*
*	Function:	_dev_init
*
*	Purpose:
*
*		Initialize the given device to its initial open() state. This is for
*		items which may change while the device is open.
*
*	Arguments:
*
*		dev			The device to initialize.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
*****************************************************************************/

static U32 _dev_init(dev_data_t* dev)
{
	U32	sid;
	U32	stat;
	U32	status	= (U32) GSC_SUCCESS;
	U32	rx_sem;
	U32	tx_sem;

	stat	= _dev_reset(dev);
	status	= (status == (U32) GSC_SUCCESS) ? stat : status;

	stat	= gsc_sem_lock(&dev->tx.gio.sem);
	tx_sem	= stat == (U32) GSC_SUCCESS;
	status	= (status == (U32) GSC_SUCCESS) ? stat : status;

	stat	= gsc_sem_lock(&dev->rx.gio.sem);
	rx_sem	= stat == (U32) GSC_SUCCESS;
	status	= (status == (U32) GSC_SUCCESS) ? stat : status;

	dev->favor_tx		= HPDI32_MISC_FAVOR_TX_DEFAULT;
	dev->strict_args	= HPDI32_MISC_STRICT_ARGUMENTS_DEFAULT;
	dev->strict_config	= HPDI32_MISC_STRICT_CONFIG_DEFAULT;

	if (dev->plx_regs.ptr == NULL)
	{
		stat	= gsc_bar_map(&dev->plx_regs);
		status	= (status == (U32) GSC_SUCCESS) ? stat : status;
	}

	if (dev->gsc_regs.ptr == NULL)
	{
		stat	= gsc_bar_map(&dev->gsc_regs);
		status	= (status == (U32) GSC_SUCCESS) ? stat : status;
	}

	// Cache some information.
	stat	= GSC_REG_READ(dev, HPDI32_FRR, &dev->cache.frr);
	status	= (status == (U32) GSC_SUCCESS) ? stat : status;
	stat	= GSC_REG_READ(dev, HPDI32_FSR, &dev->cache.fsr);
	status	= (status == (U32) GSC_SUCCESS) ? stat : status;
	stat	= GSC_REG_READ(dev, HPDI32_TFSR, &dev->cache.tfsr);
	status	= (status == (U32) GSC_SUCCESS) ? stat : status;
	stat	= GSC_REG_READ(dev, HPDI32_RFSR, &dev->cache.rfsr);
	status	= (status == (U32) GSC_SUCCESS) ? stat : status;

	if ((dev->cache.frr & HPDI32_FRR_FSR) == 0)
		dev->cache.fsr	= 0;

	if ((dev->cache.fsr & HPDI32_FSR_TFSR_RFSR) == 0)
	{
		dev->cache.tfsr	= 0;
		dev->cache.rfsr	= 0;
	}

	if ((dev->cache.fsr & HPDI32_FSR_DMDMA_CH1) == 0)
	{
		// DMA on channel 1 is not supported.
		dev->dma.ch1.capabilities	= 0;
	}

	if ((dev->cache.fsr & HPDI32_FSR_GPIO6_TXAS) &&
		(HPDI32_TX_AUTO_STOP_DEFAULT == HPDI32_TX_AUTO_STOP_NO))
	{
		// Enable the Tx Auto Stop feature
		stat	= GSC_REG_WRITE(dev, HPDI32_BCR, HPDI32_BCR_TX_AUTO_STOP_NO);
		status	= (status == (U32) GSC_SUCCESS) ? stat : status;
	}

	if (dev->cache.fsr & HPDI32_FSR_SCD)
		dev->cache.single_cycle_disable	= 1;
	else
		dev->cache.single_cycle_disable	= 0;

	stat	= gsc_reg_pci_read(&dev->device, GSC_PCI_9080_SID, &sid);
	status	= (status == (U32) GSC_SUCCESS) ? stat : status;

	if (sid == HPDI32_SUBSYSTEM_ID_64)
		dev->cache.bus_width	= 64;
	else
		dev->cache.bus_width	= 32;

	if (dev->cache.bus_width == 64)
	{
		// The User Jumpers are supported on all 64-bit boards.
		dev->cache.user_jumpers	= 1;
	}
	else if (dev->cache.frr & HPDI32_FRR_CONFORMANT)
	{
		// User Jumpers are unsupported on 32-bit PMC boards.

		if (HPDI32_FRR_SPEC_FF_DECODE(dev->cache.frr) == HPDI32_FRR_SPEC_FF_PMC)
			dev->cache.user_jumpers	= 0;
		else
			dev->cache.user_jumpers	= 1;
	}
	else if (dev->cache.frr & HPDI32_FRR_PMC)
	{
		// User Jumpers are unsupported on 32-bit PMC boards.
		dev->cache.user_jumpers	= 0;
	}
	else
	{
		dev->cache.user_jumpers	= 1;
	}

	// Other items.
	stat	= io_dma_init(dev);
	status	= (status == (U32) GSC_SUCCESS) ? stat : status;

	stat	= io_init(dev, &dev->rx);
	status	= (status == (U32) GSC_SUCCESS) ? stat : status;
	stat	= io_init(dev, &dev->tx);
	status	= (status == (U32) GSC_SUCCESS) ? stat : status;

	stat	= irq_init(dev);
	status	= (status == (U32) GSC_SUCCESS) ? stat : status;

	stat	= gsc_irq_enable(&dev->irq.girq);
	status	= (status == (U32) GSC_SUCCESS) ? stat : status;

	if (rx_sem)
		gsc_sem_unlock(&dev->rx.gio.sem);

	if (tx_sem)
		gsc_sem_unlock(&dev->tx.gio.sem);

	return(status);
}



/*****************************************************************************
*
*	Function:	_dev_list_access_lock
*
*	Purpose:
*
*		Get exclusive access to our device list.
*
*	Arguments:
*
*		None.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
*****************************************************************************/

static U32 _dev_list_access_lock(void)
{
	U32	status;

	status	= gsc_sem_lock(&_dev_sem);
	return(status);
}



/*****************************************************************************
*
*	Function:	_dev_list_access_unlock
*
*	Purpose:
*
*		Release exclusive access to our device list.
*
*	Arguments:
*
*		None.
*
*	Returned:
*
*		None.
*
*****************************************************************************/

static void _dev_list_access_unlock(void)
{
	gsc_sem_unlock(&_dev_sem);
}



/*****************************************************************************
*
*	Function:	_dev_list_index_acquire
*
*	Purpose:
*
*		Locate an unused device list entry.
*
*	Arguments:
*
*		index		The index obtained is recorded here.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
*****************************************************************************/

static U32 _dev_list_index_acquire(int* index)
{
	int	i;
	U32	status	= (U32) GSC_TOO_MANY_OPEN_HANDLES;

	for (i = 0; i < (int) ARRAY_ELEMENTS(_dev_list); i++)
	{
		if (GSC_DEVICE_IS_OPEN(&_dev_list[i].device))
			continue;

		status		= (U32) GSC_SUCCESS;
		index[0]	= i;
		break;
	}

	return(status);
}



/*****************************************************************************
*
*	Function:	dev_access_library_begin
*
*	Purpose:
*
*		Initialize the library prior to being used.
*
*	Arguments:
*
*		None.
*
*	Returned:
*
*		None.
*
*****************************************************************************/

void dev_access_library_begin(void)
{
	api_initialized	= TRUE;
	memset(_dev_list, 0, sizeof(_dev_list));

	api_status	= gsc_sem_create(&_dev_sem, "DEV LIST");

	if (api_status == (U32) GSC_SUCCESS)
		gsc_compute_base_index();

	if (api_status == (U32) GSC_SUCCESS)
		api_status	= dev_version_verify();

	if (api_status == (U32) GSC_SUCCESS)
		api_status	= config_init();

	if (api_status == (U32) GSC_SUCCESS)
		api_status	= gsc_drv_init();
}



/*****************************************************************************
*
*	Function:	dev_access_library_end
*
*	Purpose:
*
*		Release all resources as the library is being unloaded.
*
*	Arguments:
*
*		None.
*
*	Returned:
*
*		None.
*
*****************************************************************************/

void dev_access_library_end(void)
{
	int	i;

	if (api_initialized)
	{
		// Close all devices and unmap PLX and GSC registers.

		for (i = 0; i < (int) ARRAY_ELEMENTS(_dev_list); i++)
			hpdi32_close(GSC_GET_HANDLE(&_dev_list[i].device));

		// End library usage.
		gsc_sem_lock(&_dev_sem);
		gsc_sem_destroy(&_dev_sem);
		api_status		= (U32) GSC_SUCCESS;
		api_initialized	= 0;
	}
}



/*****************************************************************************
*
*	Function:	dev_access_lock
*
*	Purpose:
*
*		Validate the application handle, and if valid, acquire the semaphore
*		for the device and provide the caller with a pointer to the device
*		structure.
*
*	Arguments:
*
*		handle		The handle to validate.
*
*		dev			The structure pointer is recorded here.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
*****************************************************************************/

U32 dev_access_lock(void* handle, dev_data_t** dev)
{
	U32	status;

	for (;;)	/* A convenience loop. */
	{
		dev[0]	= NULL;

		if (api_status != (U32) GSC_SUCCESS)
		{
			status	= (U32) GSC_INIT_FAILURE;
			break;
		}

		status	= _dev_list_access_lock();

		if (status != (U32) GSC_SUCCESS)
			break;

		status	= _app_handle_validate(handle, dev);

		if (status != (U32) GSC_SUCCESS)
		{
			_dev_list_access_unlock();
			break;
		}

		status	= _dev_access_lock(dev[0]);
		_dev_list_access_unlock();
		break;
	}

	return(status);
}



/*****************************************************************************
*
*	Function:	dev_access_unlock
*
*	Purpose:
*
*		Release exclusive access to the given device.
*
*	Arguments:
*
*		dev		The device to unlock.
*
*	Returned:
*
*		None.
*
*****************************************************************************/

void dev_access_unlock(dev_data_t* dev)
{
	gsc_sem_unlock(&dev->sem);
}



/*****************************************************************************
*
*	Function:	hpdi32_board_count
*
*	Purpose:
*
*		Count the number of installed boards.
*
*	Arguments:
*
*		count		The count is recorded here upon success.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
*****************************************************************************/

U32 GSC_EXPORT hpdi32_board_count(U8* count)
{
	U32	status;

	if (api_status != (U32) GSC_SUCCESS)
		status	= (U32) GSC_INIT_FAILURE;
	else if (count == NULL)
		status	= (U32) GSC_NULL_PARAM;
	else
		status	= gsc_count_boards(count);

	return(status);
}



/*****************************************************************************
*
*	Function:	hpdi32_close
*
*	Purpose:
*
*		Close an open communications channel.
*
*	Arguments:
*
*		handle		The handle of the channel to close.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
*****************************************************************************/

U32 GSC_EXPORT hpdi32_close(void* handle)
{
	dev_data_t*	dev;
	U32			status;

	for (;;)	/* A convenience loop. */
	{
		if (api_status != (U32) GSC_SUCCESS)
		{
			status	= (U32) GSC_INIT_FAILURE;
			break;
		}

		status	= _dev_list_access_lock();

		if (status != (U32) GSC_SUCCESS)
			break;

		status	= _app_handle_validate(handle, &dev);

		if (status != (U32) GSC_SUCCESS)
		{
			_dev_list_access_unlock();
			break;
		}

		status	= _dev_access_lock(dev);

		if (status != (U32) GSC_SUCCESS)
		{
			_dev_list_access_unlock();
			break;
		}

		// We ignore status values from here as we ARE closing the device.
		_dev_reset(dev);
		gsc_irq_disable(&dev->irq.girq);
		_dev_data_destroy(dev);
		status	= gsc_close(&dev->device);
		_dev_list_access_unlock();
		break;
	}

	return(status);
}



/*****************************************************************************
*
*	Function:	hpdi32_init
*
*	Purpose:
*
*		Initialize the referenced device and our recorded settings to the
*		initial open() state.
*
*	Arguments:
*
*		handle		The handle of the device to access.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
*****************************************************************************/

U32 GSC_EXPORT hpdi32_init(void* handle)
{
	dev_data_t*	dev;
	U32			status;

	status	= dev_access_lock(handle, &dev);

	if (status == (U32) GSC_SUCCESS)
	{
		status	= _dev_init(dev);
		dev_access_unlock(dev);
	}

	return(status);
}



/*****************************************************************************
*
*	Function:	hpdi32_open
*
*	Purpose:
*
*		Open a communications channel to an HPDI32.
*
*	Arguments:
*
*		index		The index of the board to access.
*
*		handle		The resulting handle is recorded here. It is NULL if
*					open fails.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
*****************************************************************************/

U32 GSC_EXPORT hpdi32_open(U8 index, void** handle)
{
	int	lst_index;
	U32	status;

	for (;;)	/* A convenience loop. */
	{
		if (api_status != (U32) GSC_SUCCESS)
		{
			status	= (U32) GSC_INIT_FAILURE;
			break;
		}

		if (handle == NULL)
		{
			status	= (U32) GSC_NULL_PARAM;
			break;
		}

		handle[0]	= NULL;
		status		= _dev_list_access_lock();

		if (status != (U32) GSC_SUCCESS)
			break;

		status	= _dev_list_index_acquire(&lst_index);

		if (status != (U32) GSC_SUCCESS)
		{
			_dev_list_access_unlock();
			break;
		}

		status	= gsc_open(	(U8) index,
							(U8) lst_index,
							&_dev_list[lst_index].device,
							handle);

		if (status == (U32) GSC_SUCCESS)
			status	= _dev_data_create(&_dev_list[lst_index]);

		if (status == (U32) GSC_SUCCESS)
			status	= _dev_init(&_dev_list[lst_index]);

		if (status != (U32) GSC_SUCCESS)
		{
			handle[0]	= NULL;
			_dev_data_destroy(&_dev_list[lst_index]);
			gsc_close(&_dev_list[lst_index].device);
		}

		_dev_list_access_unlock();
		break;
	}

	return(status);
}



/*****************************************************************************
*
*	Function:	hpdi32_reset
*
*	Purpose:
*
*		Reset the HPDI32.
*
*	Arguments:
*
*		handle		The handle of the device to access.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
*****************************************************************************/

U32 GSC_EXPORT hpdi32_reset(void* handle)
{
	dev_data_t*	dev;
	U32			status;

	status	= dev_access_lock(handle, &dev);

	if (status == (U32) GSC_SUCCESS)
	{
		status	= _dev_reset(dev);
		dev_access_unlock(dev);
	}

	return(status);
}



/*****************************************************************************
*
*	Function:	hpdi32_api_status
*
*	Purpose:
*
*		Provide the caller with the library's recorded status.
*
*	Arguments:
*
*		stat		The recorded status is put here.
*
*		arg			Any associated data value is recorded here.
*
*		api_ver		This is the API version number the application was for.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
*****************************************************************************/

U32 GSC_EXPORT hpdi32_api_status(U32* stat, U32* arg, U32 api_ver)
{
	U32	status;

	if ((stat == NULL) || (arg == NULL))
	{
		status	= (U32) GSC_NULL_PARAM;
	}
	else
	{
		if (api_ver == HPDI32_API_VERSION)
			api_status	= (U32) GSC_SUCCESS;
		else if ((api_ver >= 2) && (api_ver <= 2))
			api_status	= (U32) GSC_SUCCESS;
		else
			api_status	= (U32) GSC_INVALID_VERSION_API;

		status	= (U32) GSC_SUCCESS;
		stat[0]	= api_status;
		arg[0]	= api_status_arg;
	}

	return(status);
}


