// API - Any Device - Any OS
// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/HPDI32/HPDI32_Linux_sdk_6.x.x_GSC_DN/api/gsc_bar.c $
// $Rev: 23674 $
// $Date: 2013-10-31 16:25:34 -0500 (Thu, 31 Oct 2013) $

#include <string.h>

#include "gsc_bar.h"

#include "gsc_drv.h"
#include "gsc_reg.h"



/*****************************************************************************
*
*	Function:	gsc_bar_create
*
*	Purpose:
*
*		Prepare the given structure for use with an open device.
*
*	Arguments:
*
*		bar			The structure to initialize.
*
*		device		The data used to access the device.
*
*		index		The Base Address Register Index.
*
*		reg			The register definition used to access the particular
*					base Address Register.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
*****************************************************************************/

U32 gsc_bar_create(gsc_bar_t* bar, gsc_device_t* device, U32 index, U32 reg)
{
	U32	status;

	memset(bar, 0, sizeof(bar[0]));
	bar->device	= device;
	bar->index	= (U8) index;
	bar->reg	= reg;

	status	= gsc_bar_init(bar);
	return(status);
}



/*****************************************************************************
*
*	Function:	gsc_bar_destroy
*
*	Purpose:
*
*		Reset the given structure as it is no longer being used.
*
*	Arguments:
*
*		handle		The handle used to access the device.
*
*		bar			The structure to reset. The "reg" and "index"
*					fields are assumed to be valid.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
*****************************************************************************/

U32 gsc_bar_destroy(gsc_bar_t* bar)
{
	U32	status;

	status	= gsc_bar_unmap(bar);
	memset(bar, 0, sizeof(bar[0]));
	return(status);
}



/*****************************************************************************
*
*	Function:	gsc_bar_init
*
*	Purpose:
*
*		Initialize the given structure to its default state.
*
*	Arguments:
*
*		bar			The structure to initialize.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
*****************************************************************************/

U32 gsc_bar_init(gsc_bar_t* bar)
{
	U32	status;
	U32	value;

	status	= gsc_reg_pci_read(bar->device, bar->reg, &value);

	if (status == (U32) GSC_SUCCESS)
	{
		status	= gsc_drv_bar_size_get(	bar->device,
										(S8) bar->index,
										&bar->size);
	}

	if (status != (U32) GSC_SUCCESS)
	{
		bar->io_space	= FALSE;
		bar->address	= 0;
		bar->ptr		= NULL;
	}
	else if (value & 0x1)
	{
		// An I/O space region.
		bar->io_space	= TRUE;
		bar->address	= value & 0xFFFFFFFC;
		bar->ptr		= NULL;
	}
	else
	{
		// A memory region.
		bar->io_space	= FALSE;
		bar->address	= value & 0xFFFFFFF0;
		bar->ptr		= NULL;
		status			= gsc_bar_map(bar);

		if (status == (U32) GSC_INVALID_ADDRESS)
		{
			// We can't do it under the current system.
			// We'll have to access things the slow way.
			status	= (U32) GSC_SUCCESS;
		}
		else if (status == (U32) GSC_INSUFFICIENT_RESOURCES)
		{
			// Not enough Page Table Entries.
			// We'll have to access things the slow way.
			status	= (U32) GSC_SUCCESS;
		}
	}

	return(status);
}



/*****************************************************************************
*
*	Function:	gsc_bar_map
*
*	Purpose:
*
*		Map the given Base Address Register region to user space.
*
*	Arguments:
*
*		bar			The structure for the Base Address Register of interest.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
*****************************************************************************/

U32 gsc_bar_map(gsc_bar_t* bar)
{
	U32	status;

	if (bar->io_space)
	{
		status	= (U32) GSC_INVALID_PCI_SPACE;
	}
	else if (bar->address == 0)
	{
		status	= (U32) GSC_INVALID_ADDRESS;
	}
	else if (bar->ptr)
	{
		status	= (U32) GSC_SUCCESS;
	}
	else
	{
		status	= gsc_drv_bar_map(bar->device, (S8) bar->index, &bar->ptr);

		if (status != (U32) GSC_SUCCESS)
			bar->ptr	= 0;
	}

	return(status);
}



/*****************************************************************************
*
*	Function:	gsc_bar_reg_mod
*
*	Purpose:
*
*		Perform a read-modify-write on a register from the referenced device
*		within the given Base Address Register area. This is applicable for
*		BAR0 registers and GSC firmware registers, except that the BAR0
*		registers are read only to the application.
*
*	Arguments:
*
*		bar			The Base Address Register structure where the register
*					resides.
*
*		reg			The register to access.
*
*		value		The value to write
*
*		mask		The set of bits to modify. Excess bits are ignored.
*					If a bit is set here, then the register bits get
*					changed.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
*****************************************************************************/

U32 gsc_bar_reg_mod(const gsc_bar_t* bar, U32 reg, U32 value, U32 mask)
{
	U32	status;

	if (bar->ptr)
	{
		// This is a directly accessible memory mapped register.
		status	= gsc_reg_mapped_mod((UL) bar->ptr, bar->size, reg, value, mask);
	}
	else if (bar->io_space)
	{
		// This is an I/O mapped register.
		status	= gsc_reg_io_mod(	bar->device,
									bar->index,
									bar->size,
									reg,
									value,
									mask);
	}
	else
	{
		// This is a NON-directly accessible memory mapped register.
		status	= gsc_reg_mem_mod(	bar->device,
									bar->index,
									bar->size,
									reg,
									value,
									mask);
	}

	return(status);
}



/*****************************************************************************
*
*	Function:	gsc_bar_reg_read
*
*	Purpose:
*
*		Read a register from the referenced device within the given Base
*		Address Register area. This is applicable for BAR0 registers and GSC
*		firmware registers.
*
*	Arguments:
*
*		bar			The Base Address Register structure where the register
*					resides.
*
*		reg			The register to access.
*
*		value		The value read is recorded here.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
*****************************************************************************/

U32 gsc_bar_reg_read(const gsc_bar_t* bar, U32 reg, U32* value)
{
	U32	status;

	if (bar->ptr)
	{
		// This is a directly accessible memory mapped register.
		status	= gsc_reg_mapped_read((UL) bar->ptr, bar->size, reg, value);
	}
	else if (bar->io_space)
	{
		// This is an I/O mapped register.
		status	= gsc_reg_io_read(	bar->device,
									bar->index,
									bar->size,
									reg,
									value);
	}
	else
	{
		// This is a NON-directly accessible memory mapped register.
		status	= gsc_reg_mem_read(	bar->device,
									bar->index,
									bar->size,
									reg,
									value);
	}

	return(status);
}



/*****************************************************************************
*
*	Function:	gsc_bar_reg_write
*
*	Purpose:
*
*		Write to a register from the referenced device within the given Base
*		Address Register area. This is applicable for BAR0 registers and GSC
*		firmware registers, except that the BAR0 registers are read only to
*		the application.
*
*	Arguments:
*
*		handle		The device to access.
*
*		bar			The Base Address Register structure where the register
*					resides.
*
*		reg			The register to access.
*
*		value		The value to write
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
*****************************************************************************/

U32 gsc_bar_reg_write(const gsc_bar_t* bar, U32 reg, U32 value)
{
	U32	status;

	if (bar->ptr)
	{
		// This is a directly accessible memory mapped register.
		status	= gsc_reg_mapped_write((UL) bar->ptr, bar->size, reg, value);
	}
	else if (bar->io_space)
	{
		// This is an I/O mapped register.
		status	= gsc_reg_io_write(	bar->device,
									bar->index,
									bar->size,
									reg,
									value);
	}
	else
	{
		// This is a NON-directly accessible memory mapped register.
		status	= gsc_reg_mem_write(bar->device,
									bar->index,
									bar->size,
									reg,
									value);
	}

	return(status);
}



/*****************************************************************************
*
*	Function:	gsc_bar_unmap
*
*	Purpose:
*
*		Unmap the given Base Address Register region from user space.
*
*	Arguments:
*
*		bar			The structure for the Base Address Register of interest.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
*****************************************************************************/

U32 gsc_bar_unmap(gsc_bar_t* bar)
{
	U32	status;

	if (bar->io_space)
	{
		status	= (U32) GSC_SUCCESS;
	}
	else if (bar->address == 0)
	{
		status	= (U32) GSC_SUCCESS;
	}
	else if (bar->ptr == NULL)
	{
		status	= (U32) GSC_SUCCESS;
	}
	else
	{
		status	= gsc_drv_bar_unmap(bar->device, (S8) bar->index, bar->ptr);

		if (status == (U32) GSC_SUCCESS)
			bar->ptr	= 0;
	}

	return(status);
}


