// API - Any Device - Linux
// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/HPDI32/HPDI32_Linux_sdk_6.x.x_GSC_DN/api/gsc_os_event.c $
// $Rev: 23674 $
// $Date: 2013-10-31 16:25:34 -0500 (Thu, 31 Oct 2013) $

#include <errno.h>
#include <string.h>
#include <sys/time.h>

#include "dev_drv.h"
#include "gsc_os_event.h"
#include "gsc_os_error.h"



/* #defines	*****************************************************************/

#define	ARRAY_ELEMENTS(a)		(sizeof((a))/sizeof((a)[0]))
#define	_EVENT_HANDLE_CREATE(e)	((unsigned long)(e))
#define	_EVENT_HANDLE_VALID(e)	(((e)) && ((e)->handle == (unsigned long) (e))) ? 1 : 0



/*****************************************************************************
*
*	Function:	_node_append
*
*	Purpose:
*
*		Append a node to the end of the list. The list must be locked when
*		we are called.
*
*	Arguments:
*
*		node	The node to add.
*
*		device	The OS specific data is stored here.
*
*	Returned:
*
*		None.
*
*****************************************************************************/

static void _node_append(os_event_node_t* node, os_device_t* device)
{
	OS_DEVICE_T_KEY_VERIFY(device);

	if ((device->events.head == (os_event_node_t*) &device->events.tail) ||
		(device->events.head == NULL))
	{
		node->prev			= (os_event_node_t*) &device->events.head;
		node->next			= (os_event_node_t*) &device->events.tail;
		device->events.head	= node;
		device->events.tail	= node;
	}
	else
	{
		node->prev			= device->events.tail;
		node->next			= (os_event_node_t*) &device->events.tail;
		node->prev->next	= node;
		device->events.tail	= node;
	}

	OS_DEVICE_T_KEY_VERIFY(device);
}



/*****************************************************************************
*
*	Function:	_node_remove
*
*	Purpose:
*
*		Remove a node from the list. The list must be locked when we are
*		called.
*
*	Arguments:
*
*		node	The node to add.
*
*		device	The OS specific data is stored here.
*
*	Returned:
*
*		None.
*
*****************************************************************************/

static void _node_remove(os_event_node_t* node, os_device_t* device)
{
	// In some cases application threads don't complete until after the
	// device is closed. In those cases the node device structure isn't
	// around to use any longer.

	if (node->trash == 0)
	{
		OS_DEVICE_T_KEY_VERIFY(device);

		if (node->prev == (os_event_node_t*) &device->events.head)
			device->events.head	= node->next;
		else
			node->prev->next	= node->next;

		if (node->next == (os_event_node_t*) &device->events.tail)
			device->events.tail	= node->prev;
		else
			node->next->prev	= node->prev;

		node->prev	= NULL;
		node->next	= NULL;

		OS_DEVICE_T_KEY_VERIFY(device);
	}
}



/*****************************************************************************
*
*	Function:	_suspend_thread
*
*	Purpose:
*
*		Suspend the current task until one of the listed events is triggerred.
*
*	Arguments:
*
*		count		The number of events being waited on.
*
*		list		The list of structures for the events we're waiting on.
*
*		timeout_ms	The wait timeout perios in milleseconds.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
*****************************************************************************/

static U32 _suspend_thread(U8 count, const os_event_t** list, U32 timeout_ms)
{
	os_device_t*	device	= list[0]->device;
	int				i;
	U32				ms;
	os_event_node_t	node;
	U32				sec;
	U32				status;
	struct timespec	ts;
	struct timeval	tv;
	struct timezone	tz;
	U32				us;

	// Acquire exclusive access to the node list.
	status	= os_sem_lock(&device->events.sem);

	if (status != GSC_SUCCESS)
	{
	}
	else if (device->events.quit)
	{
		os_sem_unlock(&device->events.sem);
	}
	else
	{
		// Initialize the node.
		memset(&node, 0, sizeof(node));
		node.count	= count;
		node.list	= list;
		pthread_cond_init(&node.cond, NULL);
		gettimeofday(&tv, &tz);
		sec			= timeout_ms / 1000;
		ms			= timeout_ms % 1000;
		us			= tv.tv_usec + (ms * 1000);

		sec			= sec + (us / 1000000L);
		us			= us % 1000000L;

		ts.tv_sec	= tv.tv_sec + sec;
		ts.tv_nsec	= us * 1000;

		// Add this node to the list.
		_node_append(&node, device);

		// Wait for the event or the timeout.
		// This call unlocks the mutex while we're blocked.
		i		= pthread_cond_timedwait(	&node.cond,
											&device->events.sem.mutex,
											&ts);

		if (i == 0)
			status	= GSC_SUCCESS;
		else if (i == ETIMEDOUT)
			status	= GSC_WAIT_TIMEOUT;
		else
			status	= os_errno_2_status(i);

		// Remove the node from the list.
		_node_remove(&node, device);

		// Cleanup.
		pthread_cond_destroy(&node.cond);

		// Release exclusive access to the node list.
		os_sem_unlock(&device->events.sem);
	}

	return(status);
}



/*****************************************************************************
*
*	Function:	os_event_create
*
*	Purpose:
*
*		Create an event structure and its content.
*
*	Arguments:
*
*		event		The OS specific data is stored here.
*
*		name		A name to associate with the event.
*
*		device		The structure for the device being accessed.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
*****************************************************************************/

U32 os_event_create(os_event_t* event, const char* name, os_device_t* device)
{
	U32	status;

	for (;;)	// A convenience loop.
	{
		if ((event == NULL) || (name == NULL) || (device == NULL))
		{
			status	= GSC_NULL_PARAM;
			break;
		}

		if (name[0] == 0)
		{
			status	= GSC_INVALID_DATA;
			break;
		}

		OS_DEVICE_T_KEY_VERIFY(device);
		memset(event, 0, sizeof(os_event_t));
		OS_KEY_SET(event, OS_EVENT_T_KEY);
		OS_EVENT_T_KEY_VERIFY(event);
		event->handle	= _EVENT_HANDLE_CREATE(event);
		event->device	= device;
		strncpy(event->name, name, sizeof(event->name));
		event->name[sizeof(event->name) - 1] = 0;
		OS_DEVICE_T_KEY_VERIFY(device);
		OS_EVENT_T_KEY_VERIFY(event);
		status	= GSC_SUCCESS;
		break;
	}

	return(status);
}



/*****************************************************************************
*
*	Function:	os_event_destroy
*
*	Purpose:
*
*		Destroy an event.
*
*	Arguments:
*
*		event		The OS specific event data is stored here.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
*****************************************************************************/

U32 os_event_destroy(os_event_t* event)
{
	U32	status;

	// Don't do key validation, as the key may be uninitialized.

	if (event == NULL)
	{
		status	= GSC_NULL_PARAM;
	}
	else
	{
		memset(event, 0, sizeof(os_event_t));
		status	= GSC_SUCCESS;
	}

	return(status);
}



/*****************************************************************************
*
*	Function:	os_event_t_init
*
*	Purpose:
*
*		Initialize an OS event structure. The structure is assumed to not be
*		in use.
*
*	Arguments:
*
*		event		The OS specific event data is stored here.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
*****************************************************************************/

U32 os_event_t_init(os_event_t* event)
{
	U32	status;

	if (event)
	{
		memset(event, 0, sizeof(os_event_t));
		OS_KEY_SET(event, OS_EVENT_T_KEY);
		OS_EVENT_T_KEY_VERIFY(event);
		status	= GSC_SUCCESS;
	}
	else
	{
		status	= GSC_NULL_PARAM;
	}

	return(status);
}



/*****************************************************************************
*
*	Function:	os_event_trigger
*
*	Purpose:
*
*		Trigger an event.
*
*	Arguments:
*
*		event		The OS specific data is stored here.
*
*	Returned:
*
*		GSC_SUCCESS	All went well (it is still valid).
*		else		An appropriate error status.
*
*****************************************************************************/

U32 os_event_trigger(os_event_t* event)
{
	int					i;
	os_event_node_t*	node;
	U32					status;
	os_event_node_t*	tail;

	for (;;)	// A convenience loop.
	{
		if (event == NULL)
		{
			status	= GSC_NULL_PARAM;
			break;
		}

		OS_EVENT_T_KEY_VERIFY(event);
		status	= os_sem_lock(&event->device->events.sem);

		if (status != GSC_SUCCESS)
			break;

		node	= (os_event_node_t*) event->device->events.head;
		tail	= (os_event_node_t*) &event->device->events.tail;

		for (; (node) && (node != tail); node = node->next)
		{
			for (i = 0; i < node->count; i++)
			{
				if (node->list[i]->handle == event->handle)
				{
					pthread_cond_broadcast(&node->cond);
					break;
				}
			}
		}

		os_sem_unlock(&event->device->events.sem);

		/*
		*	Don't validate the event as it may have been destroyed between the
		*	time it was triggerred above and the time we got down to here.
		*/

		break;
	}

	return(status);
}



/*****************************************************************************
*
*	Function:	os_event_validate
*
*	Purpose:
*
*		See if the referenced event is still valid.
*
*	Arguments:
*
*		event		The OS specific data is stored here.
*
*	Returned:
*
*		GSC_SUCCESS	All went well (it is still valid).
*		else		An appropriate error status.
*
*****************************************************************************/

U32 os_event_validate(os_event_t* event)
{
	int	i;
	U32	status;

	if (event)
	{
		i		= _EVENT_HANDLE_VALID(event);
		status	= i ? GSC_SUCCESS : GSC_FAILED;
	}
	else
	{
		status	= GSC_FAILED;
	}

	return(status);
}



/*****************************************************************************
*
*	Function:	os_event_wait
*
*	Purpose:
*
*		Wait for a single event to be triggered, or for a timeout.
*
*	Arguments:
*
*		event		The OS specific data is stored here.
*
*		timeout_ms	The desired timeout, in milleseconds.
*
*	Returned:
*
*		GSC_SUCCESS	All went well (the event wat triggerred).
*		else		An appropriate error status.
*
*****************************************************************************/

U32 os_event_wait(const os_event_t* event, U32 timeout_ms)
{
	const os_event_t	*list[1];
	U32					status;

	if (event)
	{
		OS_EVENT_T_KEY_VERIFY(event);
		list[0]	= event;
		status	= _suspend_thread(1, list, timeout_ms);
		// Don't validate the event here as it may already have been destroyed.
	}
	else
	{
		status	= GSC_NULL_PARAM;
	}

	return(status);
}



/*****************************************************************************
*
*	Function:	os_event_wait_2
*
*	Purpose:
*
*		Wait for one of the referenced events to be triggered, or for a
*		timeout.
*
*	Arguments:
*
*		e1			The OS specific data is for event one is stored here.
*
*		e2			The OS specific data is for event two is stored here.
*
*		timeout_ms	The desired timeout, in milleseconds.
*
*	Returned:
*
*		GSC_SUCCESS	All went well (the event wat triggerred).
*		else		An appropriate error status.
*
*****************************************************************************/

U32 os_event_wait_2(const os_event_t* e1, const os_event_t* e2, U32 timeout_ms)
{
	const os_event_t	*list[2];
	U32					status;

	if ((e1) && (e2))
	{
		OS_EVENT_T_KEY_VERIFY(e1);
		OS_EVENT_T_KEY_VERIFY(e2);
		list[0]	= e1;
		list[1]	= e2;
		status	= _suspend_thread(2, list, timeout_ms);
		OS_EVENT_T_KEY_VERIFY(e1);
		OS_EVENT_T_KEY_VERIFY(e2);
	}
	else
	{
		status	= GSC_NULL_PARAM;
	}

	return(status);
}



/*****************************************************************************
*
*	Function:	os_event_wait_multi
*
*	Purpose:
*
*		Wait for one of a list of events to be triggered, or for a timeout.
*
*	Arguments:
*
*		list		The OS specific data is for event one is stored here.
*
*		count		The number of events in the list.
*
*		timeout_ms	The desired timeout, in milleseconds.
*
*	Returned:
*
*		GSC_SUCCESS	All went well (the event wat triggerred).
*		else		An appropriate error status.
*
*****************************************************************************/

U32 os_event_wait_multi(const os_event_t* list, U8 count, U32 timeout_ms)
{
	int					i;
	const os_event_t	*lst[32];
	U32					status;

	if (list == NULL)
	{
		status	= GSC_NULL_PARAM;
	}
	else if (count == 0)
	{
		status	= GSC_SUCCESS;
	}
	else if (count > ARRAY_ELEMENTS(lst))
	{
		status	= GSC_INVALID_DATA;
	}
	else
	{
		for (i = 0; i < (int) count; i++)
		{
			OS_EVENT_T_KEY_VERIFY(&list[i]);
			lst[i]	= &list[i];
		}

		status	= _suspend_thread(count, lst, timeout_ms);
	}

	return(status);
}



/*****************************************************************************
*
*	Function:	os_events_release
*
*	Purpose:
*
*		Resume all threads waiting on the given device's event list.
*
*	Arguments:
*
*		list		The OS specific data is for event one is stored here.
*
*		count		The number of events in the list.
*
*		timeout_ms	The desired timeout, in milleseconds.
*
*	Returned:
*
*		GSC_SUCCESS	All went well (the event wat triggerred).
*		else		An appropriate error status.
*
*****************************************************************************/

U32 os_events_release(os_device_t* device)
{
	int					i;
	os_event_node_t*	node;
	U32					status;
	os_event_node_t*	tail;

	if (device == NULL)
	{
		status	= GSC_NULL_PARAM;
	}
	else
	{
		device->events.quit	= 1;

		for (i = 0; i < 100; i++)
		{
			status	= os_sem_lock(&device->events.sem);

			if (status != GSC_SUCCESS)
				continue;

			node	= (os_event_node_t*) device->events.head;
			tail	= (os_event_node_t*) &device->events.tail;

			for (;;)
			{
				if ((node == NULL) || (node == tail))
					break;

				node->trash	= 1;
				pthread_cond_broadcast(&node->cond);
				node		= node->next;
			}

			os_sem_unlock(&device->events.sem);
			break;
		}
	}

	return(status);
}


