// API - Any Device - Linux
// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/HPDI32/HPDI32_Linux_sdk_6.x.x_GSC_DN/api/gsc_os_thread.c $
// $Rev: 23674 $
// $Date: 2013-10-31 16:25:34 -0500 (Thu, 31 Oct 2013) $

#include <stdlib.h>
#include <string.h>

#include "gsc_os_thread.h"
#include "gsc_os_sleep.h"


/* data types	*************************************************************/

typedef struct
{
	U32		(*function)(void* arg);
	void*	arg;
} _thread_data_t;



/*****************************************************************************
*
*	Function:	_thread_func
*
*	Purpose:
*
*		This is an OS specific thread entry function, whose purpose is to
*		call the application's OS independent thread entry function.
*
*	Arguments:
*
*		pvArg	The thread's entry arguemtn.
*
*	Returned:
*
*		The thread exit code, which we ignore.
*
*****************************************************************************/

static void* _thread_func(void* arg)
{
	_thread_data_t*	data	= (_thread_data_t*) arg;
	U32				ret;
	_thread_data_t	tmp		= data[0];

	free(arg);
	ret	= (tmp.function)(tmp.arg);
	pthread_exit((void*) (unsigned long) ret);
	return((void*) (unsigned long) ret);
}



/*****************************************************************************
*
*	Function:	os_thread_create
*
*	Purpose:
*
*		Create a thread whose effective entry point and argument are as
*		given.
*
*	Arguments:
*
*		thread		The OS specific thread data is stored here.
*
*		name		A name to associate with the thread.
*
*		function	The effective entry point.
*
*		arg			The entry argument.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
*****************************************************************************/

U32 os_thread_create(	os_thread_t*	thread,
						const char*		name,
						U32				(*function)(void* arg),
						void*			arg)
{
	_thread_data_t*	data;
	int				i;
	U32				status;

	for (;;)	// A convenience loop.
	{
		if ((thread == NULL) || (name == NULL) || (function == NULL))
		{
			status	= GSC_NULL_PARAM;
			break;
		}

		memset(thread, 0, sizeof(os_thread_t));
		OS_KEY_SET(thread, OS_THREAD_T_KEY);
		OS_THREAD_T_KEY_VERIFY(thread);
		data	= malloc(sizeof(_thread_data_t));

		if (data == NULL)
		{
			status	= GSC_INSUFFICIENT_RESOURCES;
			break;
		}

		data->function	= function;
		data->arg		= arg;
		i				= pthread_create(	&thread->thread,
											NULL,
											_thread_func,
											data);

		if (i)
		{
			free(data);
			status	= GSC_THREAD_FAILURE;
			break;
		}

		strncpy(thread->name, name, sizeof(thread->name));
		thread->name[sizeof(thread->name) - 1]	= 0;
		status	= GSC_SUCCESS;
		break;
	}

	return(status);
}



/*****************************************************************************
*
*	Function:	os_thread_destroy
*
*	Purpose:
*
*		End a thread in a well behaved manner.
*
*	Arguments:
*
*		thread		The thread to end.
*
*	Returned:
*
*		GSC_SUCCESS	All went well.
*		else		An appropriate error status.
*
*****************************************************************************/

U32 os_thread_destroy(os_thread_t* thread)
{
	U32	status;

	if (thread)
	{
		// Don't do key validation as the key may not have been initialized.
		pthread_join(thread->thread, NULL);
		memset(thread, 0, sizeof(os_thread_t));
		status	= GSC_SUCCESS;
	}
	else
	{
		status	= GSC_NULL_PARAM;
	}

	return(status);
}



/*****************************************************************************
*
*	Function:	os_thread_wait
*
*	Purpose:
*
*		Wait for a thread to run, or for a timeout.
*
*	Arguments:
*
*		thread		The OS specific data is stored here.
*
*		timeout_ms	The desired timeout, in milleseconds.
*
*	Returned:
*
*		GSC_SUCCESS	All went well (the event wat triggerred).
*		else		An appropriate error status.
*
*****************************************************************************/

U32 os_thread_wait(const os_thread_t* thread, U32 timeout_ms)
{
	// Linux doesn't seem to have a practical means for one thread to check
	// on the status of another. We could read /proc/<thread>/stat. but that
	// is way too slow. So we'll just sleep for the timeout period and hope
	// the referenced thread will run during that time. :)

	#define	_1M	(1L * 1000L * 1000L)

	if (timeout_ms == 0)
	{
		os_sleep_ms(1);
	}
	else if (timeout_ms < _1M)
	{
			os_sleep_ms(timeout_ms);
	}
	else
	{
		for (; timeout_ms > _1M; timeout_ms -= _1M)
			os_sleep_ms(_1M);

		if (timeout_ms > 0)
			os_sleep_ms(timeout_ms);
	}

	return(GSC_SUCCESS);
}


