// This software is covered by the GNU GENERAL PUBLIC LICENSE (GPL).
// Any Device - Driver - Linux
// $URL: http://subversion:8080/svn/gsc/trunk/drivers/LINUX/HPDI32/HPDI32_Linux_sdk_6.x.x_GSC_DN/driver/mem.c $
// $Rev: 23671 $
// $Date: 2013-10-31 16:23:27 -0500 (Thu, 31 Oct 2013) $

#include "main.h"


/******************************************************************************
*
*	Function:	_order_to_size
*
*	Purpose:
*
*		Convert an order value to its corresponding size in page sized units.
*
*	Arguments:
*
*		order	The page size order to convert.
*
*	Returned:
*
*		>= 0	The order corresponding to the given size.
*
******************************************************************************/

u32 _order_to_size(int order)
{
	u32	size	= PAGE_SIZE;

	// Limit sizes to 1G.

	for (; order > 0; order--)
	{
		if (size >= 0x40000000)
			break;

		size	*= 2;
	}

	return(size);
}



/******************************************************************************
*
*	Function:	_size_to_order
*
*	Purpose:
*
*		Convert a size value to a corresponding page size order value.
*
*	Arguments:
*
*		size	The desired size in bytes.
*
*	Returned:
*
*		>= 0	The order corresponding to the given size.
*
******************************************************************************/

int _size_to_order(u32 size)
{
	u32	bytes	= PAGE_SIZE;
	int	order	= 0;

	// Limit sizes to 1G.

	for (; bytes < 0x40000000;)
	{
		if (bytes >= size)
			break;

		order++;
		bytes	*= 2;
	}

	return(order);
}



/******************************************************************************
*
*	Function:	_mem_alloc
*
*	Purpose:
*
*		Allocate a block of memory pages.
*
*	Arguments:
*
*		mem		The structure for the desired allocation.
*
*		size	The desired size in bytes.
*
*	Returned:
*
*		0		All went well.
*		<0		An error code for the problem seen.
*
******************************************************************************/

static int _mem_alloc(dev_mem_t* mem, u32 size)
{
	unsigned long	adrs;
	int				order;
	int				status;

	order	= _size_to_order(size);
	size	= _order_to_size(order);

	for (;;)
	{
		adrs	= __get_free_pages(GFP_KERNEL | __GFP_NOWARN, order);

		if (adrs)
		{
			mem->krn_adrs	= adrs;
			mem->phy_adrs	= virt_to_bus((void*) adrs);
			mem->order		= order;
			mem->size		= size;
			mem->mmap_size	= size;
			status			= 0;

			for (; size >= PAGE_SIZE; size -= PAGE_SIZE, adrs += PAGE_SIZE)
				PAGE_RESERVE(adrs);

			break;
		}

		if (order == 0)
		{
			status	= -ENOMEM;
			break;
		}

		order--;
		size	/= 2;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	_mem_free
*
*	Purpose:
*
*		Free a previously allocated block of memory.
*
*	Arguments:
*
*		mem		The structure for the memory to free.
*
*	Returned:
*
*		None.
*
******************************************************************************/

static void _mem_free(dev_mem_t* mem)
{
	unsigned long	adrs	= mem->krn_adrs;
	u32				size	= mem->size;

	// 2.2	YES
	// 2.4	NO

	for (; size >= PAGE_SIZE; size -= PAGE_SIZE, adrs += PAGE_SIZE)
		PAGE_UNRESERVE(adrs);

	free_pages(mem->krn_adrs, mem->order);
	mem->index		= 0;
	mem->krn_adrs	= 0;
	mem->phy_adrs	= 0;
	mem->order		= 0;
	mem->size		= 0;
	mem->mmap_size	= 0;
}



/******************************************************************************
*
*	Function:	_mem_t_acquire
*
*	Purpose:
*
*		Obtain a new dev_mem_t structure for the PID structure.
*
*	Arguments:
*
*		pid	The structre for the calling process.
*
*		mem	The alloted structure pointer is put here.
*
*	Returned:
*
*		0	All went well.
*		<0	An error code for the problem seen.
*
******************************************************************************/

static int _mem_t_acquire(dev_pid_t* pid, dev_mem_t** mem)
{
	int			i;
	dev_mem_t*	mt;
	int			status;

	for (;;)	// A convenience loop.
	{
		// Find an available entry.

		for (i = 0; i < ARRAY_ELEMENTS(pid->mem); i++)
		{
			if (pid->mem[i] == NULL)
				break;
		}

		if (i >= ARRAY_ELEMENTS(pid->mem))
		{
			status	= -EBUSY;
			break;
		}

		// Allocate the structure.
		mt	= kmalloc(sizeof(dev_mem_t), GFP_KERNEL);

		if (mt == NULL)
		{
			status	= -ENOMEM;
			break;
		}

		pid->mem[i]	= mt;
		mem[0]		= mt;
		memset(mt, 0, sizeof(dev_mem_t));
		mt->index	= i;
		status		= 0;
		break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	_mem_t_locate
*
*	Purpose:
*
*		Locate the indicated memory structure.
*
*	Arguments:
*
*		pid		The structre for the calling process.
*
*		index	The index of the desired structure.
*
*	Returned:
*
*		NULL	It doesn't exist.
*		else	A pointer to the indicated structure.
*
******************************************************************************/

static dev_mem_t* _mem_t_locate(dev_pid_t* pid, int index)
{
	dev_mem_t*	mem;

	if ((index < 0) ||
		(index >= ARRAY_ELEMENTS(pid->mem)) ||
		(pid->mem[index] == NULL))
	{
		mem	= NULL;
	}
	else
	{
		mem	= pid->mem[index];
	}

	return(mem);
}



/******************************************************************************
*
*	Function:	_mem_t_release
*
*	Purpose:
*
*		Release the given dev_mem_t structure from the PID structure.
*
*	Arguments:
*
*		pid	The structre for the calling process.
*
*		mem	The structure to release.
*
*	Returned:
*
*		0	All went well.
*		< 0	An appropriate error status.
*
******************************************************************************/

static int _mem_t_release(dev_pid_t* pid, dev_mem_t* mem)
{
	int	i;
	int	status	= -EINVAL;

	for (i = ARRAY_ELEMENTS(pid->mem); i--;)
	{
		if (pid->mem[i] == mem)
		{
			status	= 0;
			kfree(mem);
			pid->mem[i]	= NULL;
			break;
		}
	}

	return(status);
}



/******************************************************************************
*
*	Function:	mem_alloc_ioctl
*
*	Purpose:
*
*		Allocate a block of physically continuous memory.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		mem	The data for the memory request.
*
*	Returned:
*
*		0	All went well.
*		<0	An error code for the problem seen.
*
******************************************************************************/

int mem_alloc_ioctl(dev_data_t* dev, gsc_drv_mem_t* mem)
{
	unsigned long	adrs_limit;
	dev_mem_t*		mt;
	dev_pid_t*		pid;
	int				status;

	for (;;)	// A convenience loop.
	{
		pid	= pid_locate(dev);

		if (pid == NULL)
		{
			status		= -ENODEV;
			mem->status	= GSC_INVALID_DEVICE_INFO;
			break;
		}

		if (mem->size == 0)
		{
			status				= 0;
			mem->status			= GSC_SUCCESS;
			mem->index			= 0;
			mem->mmap_offset	= 0;
			mem->dma_address	= 0;
			break;
		}

		status	= _mem_t_acquire(pid, &mt);

		if (status < 0)
		{
			mem->status	= GSC_INSUFFICIENT_RESOURCES;
			break;
		}

		status	= _mem_alloc(mt, mem->size);

		if (status)
		{
			mem->status	= GSC_INSUFFICIENT_RESOURCES;
			_mem_t_release(pid, mt);
			break;
		}

		status	= mmap_mem_compute_settings(dev, pid, mt);

		if (status)
		{
			mem->status	= GSC_INSUFFICIENT_RESOURCES;
			_mem_free(mt);
			_mem_t_release(pid, mt);
			break;
		}

		adrs_limit	= 0xFFFFFFFFUL - mt->size + 1;

		if (mt->phy_adrs > adrs_limit)
		{
			// We only support 32-bit address ranges.
			mem->status	= GSC_INVALID_ADDRESS;
			_mem_free(mt);
			_mem_t_release(pid, mt);
			break;
		}

		mem->status			= GSC_SUCCESS;
		mem->index			= mt->index;
		mem->size			= mt->size;
		mem->mmap_offset	= mt->mmap_offset;
		mem->dma_address	= mt->phy_adrs;
		break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	mem_free_ioctl
*
*	Purpose:
*
*		Free a previously allocated block of physically continuous memory.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		mem	The data for the memory to free.
*
*	Returned:
*
*		0	All went well.
*		<0	An error code for the problem seen.
*
******************************************************************************/

int mem_free_ioctl(dev_data_t* dev, gsc_drv_mem_t* mem)
{
	int			i;
	dev_mem_t*	mt;
	dev_pid_t*	pid;
	int			status;

	for (;;)	// A convenience loop.
	{
		pid	= pid_locate(dev);

		if (pid == NULL)
		{
			status		= -ENODEV;
			mem->status	= GSC_INVALID_DEVICE_INFO;
			break;
		}

		mt	= _mem_t_locate(pid, mem->index);

		if (mt == NULL)
		{
			status		= -ENODEV;
			mem->status	= GSC_INVALID_DATA;
			break;
		}

		if (mt->mmap_users)
		{
			status		= -EBUSY;
			mem->status	= GSC_ACCESS_DENIED;
			break;
		}

		i	= mem->index;
		_mem_free(mt);
		_mem_t_release(pid, mt);
		pid->mem[i]	= NULL;
		status	= 0;
		memset(mem, 0, sizeof(gsc_drv_mem_t));
		mem->status	= GSC_SUCCESS;
		break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	mem_info_ioctl
*
*	Purpose:
*
*		Retrieve information about an allocated memory block.
*
*	Arguments:
*
*		dev	The data for the device of interest.
*
*		mem	The data for the memory block of interest.
*
*	Returned:
*
*		0	All went well.
*		<0	An error code for the problem seen.
*
******************************************************************************/

int mem_info_ioctl(dev_data_t* dev, gsc_drv_mem_t* mem)
{
	dev_mem_t*	mt;
	dev_pid_t*	pid;
	int			status;

	for (;;)	// A convenience loop.
	{
		pid	= pid_locate(dev);

		if (pid == NULL)
		{
			status		= -ENODEV;
			mem->status	= GSC_INVALID_DEVICE_INFO;
			break;
		}

		status		= 0;
		mem->status	= GSC_SUCCESS;
		mt			= _mem_t_locate(pid, mem->index);

		if (mt == NULL)
		{
			mem->size			= 0;
			mem->mmapped		= 0;
			mem->mmap_offset	= 0;
			mem->dma_address	= 0;
		}
		else
		{
			mem->size			= mt->size;
			mem->mmapped		= mt->mmap_users ? 1 : 0;
			mem->mmap_offset	= mt->mmap_offset;
			mem->dma_address	= mt->phy_adrs;
		}

		break;
	}

	return(status);
}



/******************************************************************************
*
*	Function:	mem_dma_locate
*
*	Purpose:
*
*		Locate a memory structure matching the given settings. For DMA
*		purposes, the given address need only fall within an allocated
*		region.
*
*	Arguments:
*
*		pid		The data for the requesting process.
*
*		address	The address given to the DMA request. This is the physical
*				address on the PCI bus.
*
*	Returned:
*
*		NULL	A matching region was not found.
*		else	A matching region was found.
*
******************************************************************************/

dev_mem_t* mem_dma_locate(dev_pid_t* pid, u32 address)
{
	u32			end;
	int			i;
	dev_mem_t*	mem	= NULL;

	for (i = 0; i < ARRAY_ELEMENTS(pid->mem); i++)
	{
		if (pid->mem[i] == NULL)
			continue;

		if (pid->mem[i]->size == 0)
			continue;

		if (pid->mem[i]->phy_adrs == 0)
			continue;

		if (address < pid->mem[i]->phy_adrs)
			continue;

		end	= pid->mem[i]->phy_adrs + pid->mem[i]->size - 1;

		if (address > end)
			continue;

		mem	= pid->mem[i];
		break;
	}

	return(mem);
}



/******************************************************************************
*
*	Function:	mem_mmap_locate
*
*	Purpose:
*
*		Locate a memory structure matching the given settings.
*
*	Arguments:
*
*		pid		The data for the requesting process.
*
*		offset	The offset for the region of interest.
*
*		size	The size of the region of interest.
*
*		exact	Does the size have to match exactly? It not, then it can be
*				smaller that the listed region.
*
*	Returned:
*
*		NULL	A matching region was not found.
*		else	A matching region was found.
*
******************************************************************************/

dev_mem_t* mem_mmap_locate(dev_pid_t* pid, u32 offset, u32 size, int exact)
{
	int			i;
	dev_mem_t*	mem	= NULL;

	for (i = 0; i < ARRAY_ELEMENTS(pid->mem); i++)
	{
		if (pid->mem[i] == NULL)
			continue;

		if (pid->mem[i]->size == 0)
			continue;

		if (pid->mem[i]->mmap_offset == 0)
			continue;

		if (offset != pid->mem[i]->mmap_offset)
			continue;

		if ((exact) && (size != pid->mem[i]->size))
			continue;

		if ((exact == 0) && (size > pid->mem[i]->size))
			continue;

		mem	= pid->mem[i];
		break;
	}

	return(mem);
}



